/* Copyright (C) 1997-2005, The Martin D. Adamiker LLC. All Rights Reserved. */

/*
 * Disclaimer of Warranty
 *
 * These software programs are available to the user without any license fee or
 * royalty on an "as is" basis. The Martin D. Adamiker LLC disclaims
 * any and all warranties, whether express, implied, or statuary, including any
 * implied warranties or merchantability or of fitness for a particular
 * purpose.  In no event shall the copyright-holder be liable for any
 * incidental, punitive, or consequential damages of any kind whatsoever
 * arising from the use of these programs.
 *
 * This disclaimer of warranty extends to the user of these programs and user's
 * customers, employees, agents, transferees, successors, and assigns.
 *
 * The Martin D. Adamiker LLC does not represent or warrant that the
 * programs furnished hereunder are free of infringement of any third-party
 * patents.
 */

/*
TGA2YUV.C

  To convert a TGA file into a YUV file */

#include <stdio.h>
#include <stdlib.h>

typedef enum tcmap_type
{
  nopalette = 0,
  palette   = 1
};

typedef enum timage_type
{
  uncompr_palette =  1,
  uncompr_rgb     =  2,
  compr_palette   =  9,
  compr_rgb       = 10
};

#pragma pack(1)
typedef struct
{
  unsigned char      ident_len;
  unsigned char      cmap_type;
  unsigned char      image_type;
  unsigned short int cmap_origin;
  unsigned short int cmap_len;
  unsigned char      centry_size;
  unsigned short int image_xorg;
  unsigned short int image_yorg;
  unsigned short int image_width;
  unsigned short int image_height;
  unsigned char      pixel_size;
  struct
  {
    unsigned char AttBits :4;  // 4 bits, number of attribute bits per pixel
    unsigned char Rsrvd   :1;  // 1 bit, reserved
    unsigned char OrgBit  :1;  // 1 bit, origin: 0=lower left, 1=upper left
    unsigned char IntrLve :2;  // 2 bits, interleaving flag
  } image_discr;
} TGAFILEHEADER;

typedef struct
{
  unsigned char rgbBlue; 
  unsigned char rgbGreen; 
  unsigned char rgbRed; 
  unsigned char rgbReserved; 
} RGBQUAD; 
#pragma pack()

/*
Global data:

gl_pHeader             TGAFILEHEADER * Pointer to header data
gl_pData               unsigned char * Pointer to bitmap data
gl_dwWidthDWORDAligned unsigned long   Line length including
                                       padding bytes
gl_uRed                unsigned int    color bucket
gl_uGreen              unsigned int    color bucket
gl_uBlue               unsigned int    color bucket
*/

TGAFILEHEADER *gl_pHeader;
unsigned char *gl_pData;
unsigned long  gl_dwWidthDWORDAligned;
unsigned int   gl_uRed;
unsigned int   gl_uGreen;
unsigned int   gl_uBlue;

typedef struct
{
  double x;
  double y;
  double z;
} vector;

void product(vector *pV1,
             vector *pV2,
             vector *pV3)
{
  pV3->x = pV1->y * pV2->z + pV1->z * pV2->y;
  pV3->y = pV1->z * pV2->x + pV1->x * pV2->z;
  pV3->z = pV1->x * pV2->y + pV1->y * pV2->x;
}

/*
int ReadTGAFile(const char *szFilename)

Parameters:
szFilename   const char*     Name of input file

Return value:
0 if an error occurred
1 if no error occurred

Global variables affected:
gl_pHeader
gl_lWidthDWORDAligned
gl_pData
*/
int ReadTGAFile(const char *szFilename)
{
  long lWidthInBytes;
  unsigned char *pPixels;
  long lResult;
  int y;

  FILE *pInputfile;
  pInputfile = fopen(szFilename, "rb");
  if(!pInputfile)
  {
    return 0;
  }
  gl_pHeader = (TGAFILEHEADER*)malloc(sizeof(TGAFILEHEADER));
  if(fread(gl_pHeader,
           1,
           sizeof(TGAFILEHEADER),
           pInputfile) != sizeof(TGAFILEHEADER))
  {
    return 0;
  }
  if(gl_pHeader->cmap_type != nopalette ||
     gl_pHeader->image_type != uncompr_rgb)
  {
    return 0;
  }

  // Allocate memory for DIB
  // DIB lines are DWORD aligned in memory

  lWidthInBytes = 3 * gl_pHeader->image_width;
  gl_dwWidthDWORDAligned = ((lWidthInBytes - 1) / 4 + 1) * 4;

  gl_pData =
    (unsigned char*)malloc(gl_dwWidthDWORDAligned *
                           gl_pHeader->image_height);
  if(gl_pData == 0)
  {
    return 0;
  }

  /* Go read the bits. */

  lResult = 0L;
  if(gl_pHeader->image_discr.OrgBit)
  {
    // The OrgBit is set, see TGAFILEHEADER definition.
    // Read the first line of data into the highest line of the image.
    pPixels = gl_pData + (gl_pHeader->image_height - 1) * gl_dwWidthDWORDAligned;

    // Read top down.
    for(y = 0; y < gl_pHeader->image_height; y++)
    {
      lResult = fread(pPixels,
                      1,
                      lWidthInBytes,
                      pInputfile);
      pPixels -= gl_dwWidthDWORDAligned;
    }
  }
  else
  {
    // The OrgBit is cleared, see TGAFILEHEADER definition.
    // Read the first line of data into the lowest line of the image.
    pPixels = gl_pData;

    // Read bottom up.
    for(y = 0; y < gl_pHeader->image_height; y++)
    {
      lResult = fread(pPixels,
                      1,
                      lWidthInBytes,
                      pInputfile);
      pPixels += gl_dwWidthDWORDAligned;
    }
  }
  if(lResult != lWidthInBytes)
  {
    return 0;
  }
  fclose(pInputfile);
  return 1;
}

/*
void GetColor1(const int x,
               const int y,
               RGBQUAD* pColor)

retrieve the color of a pixel

Parameters:
x             const int     x coordinate of pixel requested
y             const int     x coordinate of pixel requested
pColor        RGBQUAD*      address of target byte triple

Return value: none

Global variables affected: none
*/
void GetColor1(const int x, const int y, RGBQUAD* pColor)
{
  unsigned char* pPos1;
  unsigned char* pPos2;
  pPos1 = gl_pData + y * gl_dwWidthDWORDAligned + x * 3;
  pPos2 = (unsigned char *)pColor;
  // Put into gl_pColor.
  *pPos2++ = *pPos1++; // BLUE.
  *pPos2++ = *pPos1++; // GREEN.
  *pPos2   = *pPos1;   // RED.
}

/*
void GetColor2(const int x,
               const int y,
               RGBQUAD* pColor)

retrieve the averaged color of a quadruple
of neighbor pixels

Parameters:
x             const int     x coordinate of pixel
                            quadruple requested
y             const int     x coordinate of pixel
                            quadruple requested
pColor        RGBQUAD*      address of target byte triple

Return value: none

Global variables affected: none
*/
void GetColor2(const int x, const int y, RGBQUAD* pColor)
{
  unsigned char* pPos1;
  unsigned char* pPos2;
  pPos1 = gl_pData + 2 * y * gl_dwWidthDWORDAligned + x * 6;
  pPos2 = (unsigned char *)pColor;
  
  /* Scan a line from left */
  gl_uBlue   = *pPos1++;
  gl_uGreen  = *pPos1++;
  gl_uRed    = *pPos1++;
  gl_uBlue  += *pPos1++;
  gl_uGreen += *pPos1++;
  gl_uRed   += *pPos1;

  pPos1 += gl_dwWidthDWORDAligned;

  /* Scan line below from right */
  gl_uRed   += *pPos1--;
  gl_uGreen += *pPos1--;
  gl_uBlue  += *pPos1--;
  gl_uRed   += *pPos1--;
  gl_uGreen += *pPos1--;
  gl_uBlue  += *pPos1--;

  // Put into gl_pColor. Divide by four.
  *pPos2++ = gl_uBlue  >> 2;
  *pPos2++ = gl_uGreen >> 2;
  *pPos2   = gl_uRed   >> 2;
}

/*
void SaveYUV1(FILE *pOutputfile)

Saves 4:2:0 YUV file (same size as input file)

Parameters:
pOutputfile   FILE*         address of FILE structure

Return value: none

Global variables affected: none
*/
void SaveYUV1(FILE *pOutputfile)
{
  unsigned char *ybuf1;
  unsigned char *ybuf2;
  unsigned char *ubuf;
  unsigned char *vbuf;
  unsigned uRow;
  unsigned char *yptr1;
  unsigned char *yptr2;
  unsigned char *uptr;
  unsigned char *vptr;
  unsigned x;
  RGBQUAD Color[4];
  RGBQUAD *pColor;
  unsigned long Y[4], U, V;
  unsigned short int uWidth;
  unsigned short int uHeight;
  int i;

  uWidth  = gl_pHeader->image_width;
  uHeight = gl_pHeader->image_height;
  ybuf1 = (unsigned char *)malloc(uWidth);
  ybuf2 = (unsigned char *)malloc(uWidth);
  ubuf  = (unsigned char *)malloc(uWidth);
  vbuf  = (unsigned char *)malloc(uWidth);
  for(uRow = 0; uRow < uHeight; uRow += 2)
  {
    yptr1 = ybuf1;
    yptr2 = ybuf2;
    for(x = 0; x < uWidth; x += 2)
    {
      GetColor1(x,     uHeight - uRow - 1, Color);
      GetColor1(x + 1, uHeight - uRow - 1, Color + 1);
      GetColor1(x,     uHeight - uRow - 2, Color + 2);
      GetColor1(x + 1, uHeight - uRow - 2, Color + 3);
      pColor = Color;
      for(i = 0; i < 4; i++)
      {
        Y[i]  = (( 19595L * pColor->rgbRed + 38469L * pColor->rgbGreen +  7471L * pColor->rgbBlue) * 219L) / 255L + 1048576L;
        pColor++;
      }
      *yptr1++ = (unsigned char)((Y[0] >> 16));
      *yptr1++ = (unsigned char)((Y[1] >> 16));
      *yptr2++ = (unsigned char)((Y[2] >> 16));
      *yptr2++ = (unsigned char)((Y[3] >> 16));
    }
    fwrite(ybuf1, uWidth, 1, pOutputfile);
    fwrite(ybuf2, uWidth, 1, pOutputfile);
  }
  for(uRow = 0; uRow < uHeight; uRow += 2)
  {
    uptr = ubuf;
    vptr = vbuf;
    for(x = 0; x < uWidth; x += 2)
    {
      GetColor1(x,     uHeight - uRow - 1, Color);
      GetColor1(x + 1, uHeight - uRow - 1, Color + 1);
      GetColor1(x,     uHeight - uRow - 1, Color + 2);
      GetColor1(x + 1, uHeight - uRow - 1, Color + 3);
      pColor = Color;
      U = 0;
      for(i = 0; i < 4; i++)
      {
        U    += ((-11056L * pColor->rgbRed - 21712L * pColor->rgbGreen + 32768L * pColor->rgbBlue) * 224L) / 255L;
        pColor++;
      }
      *uptr++  = (unsigned char)((U    >> 18) + 128); // Divide by 4, too.
    }
    fwrite(ubuf, uWidth / 2, 1, pOutputfile);
  }
  for(uRow = 0; uRow < uHeight; uRow += 2)
  {
    vptr = vbuf;
    for(x = 0; x < uWidth; x += 2)
    {
      GetColor1(x,     uHeight - uRow - 1, Color);
      GetColor1(x + 1, uHeight - uRow - 1, Color + 1);
      GetColor1(x,     uHeight - uRow - 1, Color + 2);
      GetColor1(x + 1, uHeight - uRow - 1, Color + 3);
      pColor = Color;
      V = 0;
      for(i = 0; i < 4; i++)
      {
        V    += (( 32768L * pColor->rgbRed - 27440L * pColor->rgbGreen -  5328L * pColor->rgbBlue) * 224L) / 255L;
        pColor++;
      }
      *vptr++  = (unsigned char)((V    >> 18) + 128); // Divide by 4, too.
    }
    fwrite(vbuf, uWidth / 2, 1, pOutputfile);
  }
  free(vbuf);
  free(ubuf);
  free(ybuf2);
  free(ybuf1);
}

/*
void SaveYUV2(FILE *pOutputfile)

Saves 4:2:0 YUV file (half size as input file)

Parameters:
pOutputfile   FILE*         address of FILE structure

Return value: none

Global variables affected: none
*/
void SaveYUV2(FILE *pOutputfile)
{
  unsigned char *ybuf1;
  unsigned char *ybuf2;
  unsigned char *ubuf;
  unsigned char *vbuf;
  unsigned uRow;
  unsigned char *yptr1;
  unsigned char *yptr2;
  unsigned char *uptr;
  unsigned char *vptr;
  unsigned x;
  RGBQUAD Color[4];
  RGBQUAD *pColor;
  unsigned long Y[4], U, V;
  unsigned short int uWidth;
  unsigned short int uHeight;
  int i;

  uWidth  = gl_pHeader->image_width  / 2;
  uHeight = gl_pHeader->image_height / 2;
  ybuf1 = (unsigned char *)malloc(uWidth);
  ybuf2 = (unsigned char *)malloc(uWidth);
  ubuf  = (unsigned char *)malloc(uWidth);
  vbuf  = (unsigned char *)malloc(uWidth);
  for(uRow = 0; uRow < uHeight; uRow += 2)
  {
    yptr1 = ybuf1;
    yptr2 = ybuf2;
    for(x = 0; x < uWidth; x += 2)
    {
      GetColor2(x,     uHeight - uRow - 1, Color);
      GetColor2(x + 1, uHeight - uRow - 1, Color + 1);
      GetColor2(x,     uHeight - uRow - 2, Color + 2);
      GetColor2(x + 1, uHeight - uRow - 2, Color + 3);
      pColor = Color;
      for(i = 0; i < 4; i++)
      {
        Y[i]  = (( 19595L * pColor->rgbRed + 38469L * pColor->rgbGreen +  7471L * pColor->rgbBlue) * 219L) / 255L + 1048576L;
        pColor++;
      }
      *yptr1++ = (unsigned char)((Y[0] >> 16));
      *yptr1++ = (unsigned char)((Y[1] >> 16));
      *yptr2++ = (unsigned char)((Y[2] >> 16));
      *yptr2++ = (unsigned char)((Y[3] >> 16));
    }
    fwrite(ybuf1, uWidth, 1, pOutputfile);
    fwrite(ybuf2, uWidth, 1, pOutputfile);
  }
  for(uRow = 0; uRow < uHeight; uRow += 2)
  {
    uptr = ubuf;
    vptr = vbuf;
    for(x = 0; x < uWidth; x += 2)
    {
      GetColor2(x,     uHeight - uRow - 1, Color);
      GetColor2(x + 1, uHeight - uRow - 1, Color + 1);
      GetColor2(x,     uHeight - uRow - 1, Color + 2);
      GetColor2(x + 1, uHeight - uRow - 1, Color + 3);
      pColor = Color;
      U = 0;
      for(i = 0; i < 4; i++)
      {
        U    += ((-11056L * pColor->rgbRed - 21712L * pColor->rgbGreen + 32768L * pColor->rgbBlue) * 224L) / 255L;
        pColor++;
      }
      *uptr++  = (unsigned char)((U    >> 18) + 128); // Divide by 4, too.
    }
    fwrite(ubuf, uWidth / 2, 1, pOutputfile);
  }
  for(uRow = 0; uRow < uHeight; uRow += 2)
  {
    vptr = vbuf;
    for(x = 0; x < uWidth; x += 2)
    {
      GetColor2(x,     uHeight - uRow - 1, Color);
      GetColor2(x + 1, uHeight - uRow - 1, Color + 1);
      GetColor2(x,     uHeight - uRow - 1, Color + 2);
      GetColor2(x + 1, uHeight - uRow - 1, Color + 3);
      pColor = Color;
      V = 0;
      for(i = 0; i < 4; i++)
      {
        V    += (( 32768L * pColor->rgbRed - 27440L * pColor->rgbGreen -  5328L * pColor->rgbBlue) * 224L) / 255L;
        pColor++;
      }
      *vptr++  = (unsigned char)((V    >> 18) + 128); // Divide by 4, too.
    }
    fwrite(vbuf, uWidth / 2, 1, pOutputfile);
  }
  free(vbuf);
  free(ubuf);
  free(ybuf2);
  free(ybuf1);
}

int DoIt(int argc, char *argv[])
{
  FILE *pOutputfile;
  
  if(argc != 3)
  {
    fprintf(stderr,
            "Usage: tga2yuv parameters\n"
            "\n"
            "parameters (all mandatory):\n"
            "       (input file name):  name of input file\n"
            "       (output file name): name of output file\n");
    return 8;
  }
  if(!ReadTGAFile(argv[1]))
  {
    fprintf(stderr,
            "Cannot read input file '%s'\n",
            argv[1]);
    return 7;
  }
  pOutputfile = fopen(argv[2], "wb");
  if(!pOutputfile)
  {
    fprintf(stderr,
            "Cannot open output file '%s'\n",
            argv[2]);
    return 4;
  }
/* insert a call to SaveYUV1 for sizing 1:1     */
/* insert a call to SaveYUV2 for downsizing 1:2 */
  SaveYUV1(pOutputfile);
  fclose(pOutputfile);
  return 0;
}

/* used to be main function */
int tga2yuv(int argc, char *argv[])
{
  int nResult;
  gl_pHeader = 0;
  gl_pData = 0;
  nResult = DoIt(argc, argv);
  free(gl_pHeader);
  free(gl_pData);
  return nResult;
}
